@@ -2,14 +2,14 @@
 #										#
 #  Name:       RNetCDF.R							#
 #										#
-#  Version:    1.6.2-2								#
+#  Version:    1.9-1								#
 #										#
 #  Purpose:    NetCDF interface for R.						#
 #										#
-#  Author:     Pavel Michna (michna@giub.unibe.ch)				#
-#              Milton Woods (m.woods@bom.gov.au)                                #
+#  Author:     Pavel Michna (michna@users.r-forge.r-project.org)                #
+#              Milton Woods (mwoods@users.r-forge.r-project.org)                #
 #										#
-#  Copyright:  (C) 2004-2014 Pavel Michna					#
+#  Copyright:  (C) 2004-2017 Pavel Michna					#
 #										#
 #===============================================================================#
 #										#
@@ -45,6 +45,13 @@
 #  pm       14/02/12   Corrected bug in att.delete.nc                           #
 #  pm       02/06/12   Added function read.nc()                                 #
 #  pm       16/07/12   Added packing/unpacking of data (code from mw)           #
+#  mw       21/08/14   Allow reading of character vector or scalar              #
+#  mw       05/09/14   Support reading and writing raw character arrays         #
+#  mw       08/09/14   Handle reading and writing of zero-sized arrays          #
+#  mw       24/01/16   Support conversion of timestamps to/from POSIXct         #
+#  mw       29/07/17   Replace NA in count of var.get.nc and var.put.nc so that #
+#                      corresponding dimensions are read/written to end         #
+#  mw       03/10/17   Use registered entry points for .Call functions          #
 #										#
 #===============================================================================#
 
@@ -108,15 +115,14 @@ att.copy.nc <- function(ncfile.in, variable.in, attribute, ncfile.out,
         return(invisible(NULL))
     
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_copy_att",
+    nc <- .Call(R_nc_copy_att,
                 as.integer(ncfile.in),
 		as.integer(varid.in),
 		as.integer(globflag.in),
 		as.character(attname),
                 as.integer(ncfile.out),
 		as.integer(varid.out),
-		as.integer(globflag.out),
-		PACKAGE="RNetCDF")
+		as.integer(globflag.out))
 		
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -161,12 +167,11 @@ att.delete.nc <- function(ncfile, variable, attribute)
         return(invisible(NULL))
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_delete_att",
+    nc <- .Call(R_nc_delete_att,
                 as.integer(ncfile),
 		as.integer(varid),
 		as.integer(globflag),
-		as.character(attname),
-		PACKAGE="RNetCDF")
+		as.character(attname))
     
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -212,13 +217,12 @@ att.get.nc <- function(ncfile, variable, attribute)
     ifelse(attinfo$type == "NC_CHAR", numflag <- 0, numflag <- 1);
     
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_get_att",
+    nc <- .Call(R_nc_get_att,
 	        as.integer(ncfile),
 		as.integer(varid),
 		as.character(attname),
 		as.integer(numflag),
-		as.integer(globflag),
-		PACKAGE="RNetCDF")
+		as.integer(globflag))
     
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -265,14 +269,13 @@ att.inq.nc <- function(ncfile, variable, attribute)
         return(invisible(NULL))
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_inq_att",
+    nc <- .Call(R_nc_inq_att,
         	as.integer(ncfile),
 		as.integer(varid),
 		as.character(attname),
 		as.integer(attid),
 		as.integer(nameflag),
-		as.integer(globflag),
-		PACKAGE="RNetCDF")
+		as.integer(globflag))
 
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -318,7 +321,7 @@ att.put.nc <- function(ncfile, variable, name, type, value)
     ifelse(is.numeric(value), numflag <- 1, numflag <- 0)
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_put_att",
+    nc <- .Call(R_nc_put_att,
 	        as.integer(ncfile),
 		as.integer(varid),
 		as.character(name),
@@ -326,8 +329,7 @@ att.put.nc <- function(ncfile, variable, name, type, value)
 		as.integer(length(value)),
 		as.integer(numflag),
 		as.integer(globflag),
-		value,
-		PACKAGE="RNetCDF")
+		value)
     
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -373,13 +375,12 @@ att.rename.nc <- function(ncfile, variable, attribute, newname)
         return(invisible(NULL))
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_rename_att",
+    nc <- .Call(R_nc_rename_att,
                 as.integer(ncfile),
 		as.integer(varid),
 		as.integer(globflag),
 		as.character(attname),
-                as.character(newname),
-		PACKAGE="RNetCDF")
+                as.character(newname))
     
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -396,9 +397,8 @@ close.nc <- function(con, ...)
     stopifnot(class(con) == "NetCDF") 
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_close",
-                as.integer(con),
-		PACKAGE="RNetCDF")
+    nc <- .Call(R_nc_close,
+                as.integer(con))
 
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -419,13 +419,12 @@ create.nc <- function(filename, clobber=TRUE, large=FALSE, share=FALSE,
     iprefill <- ifelse(prefill == TRUE, 1, 0)
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_create",
+    nc <- .Call(R_nc_create,
 		as.character(filename),
 		as.integer(iclobber),
                 as.integer(ilarge),
                 as.integer(ishare),
-                as.integer(iprefill),
-		PACKAGE="RNetCDF")
+                as.integer(iprefill))
 	     
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -454,12 +453,11 @@ dim.def.nc <- function(ncfile, dimname, dimlength=1, unlim=FALSE)
     ncdimlength <- ifelse(unlim  == TRUE, 0, dimlength)
     
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_def_dim",
+    nc <- .Call(R_nc_def_dim,
 		as.integer(ncfile),
 		as.character(dimname),
 		as.integer(ncdimlength),
-		as.integer(unlimflag),
-		PACKAGE="RNetCDF")
+		as.integer(unlimflag))
 
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -484,12 +482,11 @@ dim.inq.nc <- function(ncfile, dimension)
     ifelse(is.character(dimension), dimname <- dimension, dimid <- dimension)
     
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_inq_dim",
+    nc <- .Call(R_nc_inq_dim,
         	as.integer(ncfile),
 		as.integer(dimid),
 		as.character(dimname),
-		as.integer(nameflag),
-		PACKAGE="RNetCDF")
+		as.integer(nameflag))
     
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -523,13 +520,12 @@ dim.rename.nc <- function(ncfile, dimension, newname)
     ifelse(is.character(dimension), dimname <- dimension, dimid <- dimension)
     
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_rename_dim",
+    nc <- .Call(R_nc_rename_dim,
         	as.integer(ncfile),
 		as.integer(dimid),
 		as.character(dimname),
 		as.integer(nameflag),
-		as.character(newname),
-		PACKAGE="RNetCDF")
+		as.character(newname))
 
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -546,9 +542,8 @@ file.inq.nc <- function(ncfile)
     stopifnot(class(ncfile) == "NetCDF") 
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_inq_file",
-                as.integer(ncfile),
-		PACKAGE="RNetCDF")
+    nc <- .Call(R_nc_inq_file,
+                as.integer(ncfile))
 
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -582,12 +577,11 @@ open.nc <- function(con, write=FALSE, share=FALSE, prefill=TRUE, ...)
     iprefill <- ifelse(prefill == TRUE, 1, 0)
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_open",
+    nc <- .Call(R_nc_open,
 		as.character(con),
 		as.integer(iwrite),
 		as.integer(ishare),
-		as.integer(iprefill),
-		PACKAGE="RNetCDF")
+		as.integer(iprefill))
 	     
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -684,9 +678,8 @@ sync.nc <- function(ncfile)
     stopifnot(class(ncfile) == "NetCDF")
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_sync",
-                as.integer(ncfile),
-		PACKAGE="RNetCDF")
+    nc <- .Call(R_nc_sync,
+                as.integer(ncfile))
 		
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -721,7 +714,7 @@ var.def.nc <- function(ncfile, varname, vartype, dimensions)
 	if(mode(dimensions) == "numeric")
             dimids <- dimensions
 	else
-            for(i in 1:length(dimensions))
+            for(i in seq_along(dimensions))
         	try(dimids[i] <- dim.inq.nc(ncfile, dimensions[i])$id,
 	            silent=TRUE)
 
@@ -733,13 +726,12 @@ var.def.nc <- function(ncfile, varname, vartype, dimensions)
     }
     
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_def_var",
+    nc <- .Call(R_nc_def_var,
         	as.integer(ncfile),
 		as.character(varname),
 		as.character(vartype),
 		as.integer(ndims),
-		as.integer(dimids),
-		PACKAGE="RNetCDF")
+		as.integer(dimids))
 		
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -751,17 +743,18 @@ var.def.nc <- function(ncfile, varname, vartype, dimensions)
 #-------------------------------------------------------------------------------#
 
 var.get.nc <- function(ncfile, variable, start=NA, count=NA, na.mode=0,
-                       collapse=TRUE, unpack=FALSE)
+                       collapse=TRUE, unpack=FALSE, rawchar=FALSE)
 {
     #-- Check args -------------------------------------------------------------#
     stopifnot(class(ncfile) == "NetCDF")
     stopifnot(is.character(variable) || is.numeric(variable))
     stopifnot(is.numeric(start) || is.logical(start))
     stopifnot(is.numeric(count) || is.logical(count))
+    stopifnot(is.logical(collapse))
     stopifnot(is.logical(unpack))
+    stopifnot(is.logical(rawchar))
     
-    if(!any(na.mode == c(0,1,2,3)))
-        stop("Wrong na.mode", call.=FALSE)
+    stopifnot(isTRUE(na.mode %in% c(0,1,2,3))) 
 
     #-- Inquire the variable ---------------------------------------------------#
     varinfo <- try(var.inq.nc(ncfile, variable))
@@ -769,147 +762,105 @@ var.get.nc <- function(ncfile, variable, start=NA, count=NA, na.mode=0,
     if(class(varinfo) == "try-error" || is.null(varinfo))
         return(invisible(NULL))
     
+    ndims      <- varinfo$ndims
+
     #-- Get the varid as integer if necessary ----------------------------------#
     ifelse(is.character(variable), varid <- varinfo$id, varid <- variable)
 
-    #-- Get the variable dimensions if start and count contain NAs -------------#
-    if(varinfo$ndims > 0) {
-	r.start <- vector()
-	r.count <- vector()
-	for(i in 1:varinfo$ndims) {
-            r.start[i] <- 1
-	    r.count[i] <- dim.inq.nc(ncfile, varinfo$dimids[i])$length
-	}
-
-        if(any(is.na(start)) && length(start) == 1)
-	    start <- rep(NA, varinfo$ndims)
-        if(any(is.na(count)) && length(count) == 1)
-	    count <- rep(NA, varinfo$ndims)
-	
-	start[is.na(start)] <- r.start[is.na(start)]
-	count[is.na(count)] <- r.count[is.na(count)]
+    # Truncate start & count to length ndims
+    # and replace NA as described in the man page:
+    if (isTRUE(is.na(start))) {
+      start <- rep(1, ndims)
+    } else if (length(start) > ndims) {
+      start <- start[seq_len(ndims)]
     }
+    stopifnot(length(start) == ndims)
+    start[is.na(start)] <- 1
 
-    #-- Handle scalar variables ------------------------------------------------#
-    if(varinfo$ndims == 0) {
-	start[1] <- 1
-	count[1] <- 1
+    if (isTRUE(is.na(count))) {
+      count <- rep(NA, ndims)
+    } else if (length(count) > ndims) {
+      count <- count[seq_len(ndims)]
+    }
+    stopifnot(length(count) == ndims)
+    for (idim in seq_len(ndims)) {
+      if (is.na(count[idim])) {
+	diminfo <- dim.inq.nc(ncfile, varinfo$dimids[idim])
+	count[idim] <- ( diminfo$length - start[idim] + 1 )
+      }
     }
-
-    #-- Allocate memory for variable and get number of dimensions --------------#
-    varsize    <- vector()
-    varsize[1] <- prod(count)
-    varsize[2] <- count[1]                           ## Needed for character data
-    ndims      <- varinfo$ndims
 
     #-- Switch from R to C convention ------------------------------------------#
-    if(varinfo$ndims > 0) {
-	c.start <- start[length(start):1] - 1
-	c.count <- count[length(count):1]    
-    } else {
-        c.start <- 0
-	c.count <- 1
-    }
-
-    #-- Check for correct number of dimensions in start and count --------------#
-    if(varinfo$ndims > 0)
-	if(length(c.start) != ndims || length(c.count) != ndims)
-            stop("Length of start/count is not ndims", call.=FALSE)
+    c.start <- rev(start-1)
+    c.count <- rev(count)
 
     #-- C function calls -------------------------------------------------------#
-    if(varinfo$type != "NC_CHAR")
-	nc <- .Call("R_nc_get_vara_double",
+    if(varinfo$type == "NC_CHAR") {
+        nc <- .Call(R_nc_get_vara_text,
         	    as.integer(ncfile),
 		    as.integer(varid),
 		    as.integer(c.start),
 		    as.integer(c.count),
-		    as.integer(varsize),
-		    PACKAGE="RNetCDF")
-    else
-        nc <- .Call("R_nc_get_vara_text",
+		    as.integer(ndims),
+                    as.integer(rawchar))
+    } else {
+	nc <- .Call(R_nc_get_vara_double,
         	    as.integer(ncfile),
 		    as.integer(varid),
 		    as.integer(c.start),
 		    as.integer(c.count),
-		    as.integer(varsize),
-		    PACKAGE="RNetCDF")
+		    as.integer(ndims))
+    }
 
     #-- Adjust data ------------------------------------------------------------#
     if(nc$status == 0) {
 	
-	#-- Set dimensions, collapse degenerate dimensions ---------------------#
-	dimflag <- 0
-	if(ndims > 0) {
-	    count.dim <- vector()
+	#-- Convert missing value to NA if defined in NetCDF file --------------#
+        if (is.numeric(nc$data) && na.mode < 3) { 
+	  tolerance <- 1*10^-5                              ## Allow rounding error
+	  na.flag   <- 0
 
-	    if(varinfo$type != "NC_CHAR") {
-		for(i in 1:ndims)       ## R convention because ndims has R order
-		    if(collapse == TRUE) {
-                	if(count[i] > 1) {
-			    count.dim <- append(count.dim, count[i])
-			    dimflag   <- 1
-			}
-		    } else {
-                   	count.dim <- append(count.dim, count[i])
-			dimflag   <- 1
-		    }
-	    } else {
-		for(i in 2:ndims)                   ## First dim is string lenght
-		    if(collapse == TRUE) {
-                	if(count[i] > 1) {
-			    count.dim <- append(count.dim, count[i])
-			    dimflag   <- 1
-			}
-		    } else {
-                   	count.dim <- append(count.dim, count[i])
-			dimflag   <- 1
-		    }
-	    }
+	  missval.flag <- 0
+	  fillval.flag <- 0
+	  
+	  fillval <- try(att.inq.nc(ncfile, varinfo$name, "_FillValue"   ), 
+	      silent=TRUE)
+	  missval <- try(att.inq.nc(ncfile, varinfo$name, "missing_value"), 
+	      silent=TRUE)
+
+	  if(!(class(fillval) == "try-error"))
+	      if(!is.null(fillval))
+		  fillval.flag <- 1
+	  if(!(class(missval) == "try-error"))
+	      if(!is.null(missval))
+		  missval.flag <- 1
+
+	  if(na.mode == 0 && missval.flag == 1) {
+	      na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
+	      na.flag  <- 1
+	  }
+	  if(na.mode == 0 && fillval.flag == 1) {
+	      na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
+	      na.flag  <- 1
+	  }
+
+	  if(na.mode == 1 && fillval.flag == 1) {
+	      na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
+	      na.flag  <- 1
+	  }
+
+	  if(na.mode == 2 && missval.flag == 1) {
+	      na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
+	      na.flag  <- 1
+	  }
+
+	  if(na.flag == 1) {
+	      nc$data[abs(nc$data-as.numeric(na.value)) < tolerance] <- NA 
+	  }
         }
 
-	#-- Convert missing value to NA if defined in NetCDF file --------------#
-	tolerance <- 1*10^-5                              ## Allow rounding error
-	na.flag   <- 0
-
-        missval.flag <- 0
-	fillval.flag <- 0
-	
-        fillval <- try(att.inq.nc(ncfile, varinfo$name, "_FillValue"   ), 
-	    silent=TRUE)
-        missval <- try(att.inq.nc(ncfile, varinfo$name, "missing_value"), 
-	    silent=TRUE)
-
-        if(!(class(fillval) == "try-error"))
-	    if(!is.null(fillval))
-	        fillval.flag <- 1
-	if(!(class(missval) == "try-error"))
-	    if(!is.null(missval))
-		missval.flag <- 1
-
-       	if(na.mode == 0 && missval.flag == 1) {
-	    na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
-	    na.flag  <- 1
-	}
-	if(na.mode == 0 && fillval.flag == 1) {
-	    na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
-	    na.flag  <- 1
-	}
-
-       	if(na.mode == 1 && fillval.flag == 1) {
-	    na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
-	    na.flag  <- 1
-	}
-
-	if(na.mode == 2 && missval.flag == 1) {
-	    na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
-	    na.flag  <- 1
-	}
-
-        if(na.flag == 1 && na.mode != 3 && varinfo$type != "NC_CHAR")
-	    nc$data[abs(nc$data-na.value) < tolerance] <- NA 
-
 	#-- Unpack variables if requested (missing values are preserved) -------#
-        if(unpack) {
+        if(unpack && is.numeric(nc$data)) {
             offset <- try(att.inq.nc(ncfile, varinfo$name, "add_offset"),
                           silent=TRUE)
             scale  <- try(att.inq.nc(ncfile, varinfo$name, "scale_factor"),
@@ -921,9 +872,26 @@ var.get.nc <- function(ncfile, variable, start=NA, count=NA, na.mode=0,
                 nc$data      <- nc$data*scale_factor+add_offset
             }
         }
-	
+
+	#-- Set dimensions, collapse degenerate dimensions ---------------------#
+        if (is.character(nc$data) && ndims > 0) {
+          # Drop string length dimension
+          datadim <- count[-1]
+        } else {
+          datadim <- count
+        }
+        if (collapse) {
+          # Drop singleton dimensions
+          datadim <- datadim[datadim!=1]
+        }
+        if (length(datadim)<1) {
+          # For compatibility with code written for RNetCDF<=1.6.x,
+          # scalars and vectors always have a dimension attribute:
+          datadim <- length(nc$data)
+        }
+        dim(nc$data) <- datadim
+
     #-- Return object if no error ----------------------------------------------#
-	ifelse(dimflag == 0, dim(nc$data) <- (1), dim(nc$data) <- count.dim)
 	return(nc$data)
     } else
         stop(nc$errmsg, call.=FALSE)
@@ -948,12 +916,11 @@ var.inq.nc <- function(ncfile, variable)
     ifelse(is.character(variable), varname <- variable, varid <- variable)
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_inq_var",
+    nc <- .Call(R_nc_inq_var,
         	as.integer(ncfile),
 		as.integer(varid),
 		as.character(varname),
-		as.integer(nameflag),
-		PACKAGE="RNetCDF")
+		as.integer(nameflag))
 
     #-- Return object if no error ----------------------------------------------#
     if(nc$status == 0) {
@@ -981,13 +948,12 @@ var.put.nc <- function(ncfile, variable, data, start=NA, count=NA, na.mode=0,
     #-- Check args -------------------------------------------------------------#
     stopifnot(class(ncfile) == "NetCDF")
     stopifnot(is.character(variable) || is.numeric(variable))
-    stopifnot(is.numeric(data) || is.character(data) || is.logical(data))
+    stopifnot(is.numeric(data) || is.character(data) || is.raw(data) || is.logical(data))
     stopifnot(is.numeric(start) || is.logical(start))
     stopifnot(is.numeric(count) || is.logical(count))
     stopifnot(is.logical(pack))
-    
-    if(!any(na.mode == c(0,1,2)))
-        stop("Wrong na.mode", call.=FALSE)
+
+    stopifnot(isTRUE(na.mode %in% c(0,1,2))) 
 
     #-- Inquire the variable ---------------------------------------------------#
     varinfo <- try(var.inq.nc(ncfile, variable))
@@ -997,6 +963,10 @@ var.put.nc <- function(ncfile, variable, data, start=NA, count=NA, na.mode=0,
 
     ndims <- varinfo$ndims
 
+    if ((is.character(data) || is.raw(data)) && varinfo$type != "NC_CHAR") {
+        stop("R character data can only be written to NC_CHAR variable",call.=FALSE)
+    }
+
     #-- Get the varid as integer if necessary ----------------------------------#
     ifelse(is.character(variable), varid <- varinfo$id, varid <- variable)
 
@@ -1008,70 +978,84 @@ var.put.nc <- function(ncfile, variable, data, start=NA, count=NA, na.mode=0,
 	    mode(data) <- "numeric"
     }
 
-    #-- Get the data dimensions if start and count contain NAs -----------------#
-    if(varinfo$ndims == 0) {
-	start[1] <- 1
-	count[1] <- 1
+    # Truncate start & count to length ndims
+    # and replace NA as described in the man page:
+    if (isTRUE(is.na(start))) {
+      start <- rep(1, ndims)
+    } else if (length(start) > ndims) {
+      start <- start[seq_len(ndims)]
     }
-    
-    if(any(is.na(start)) && length(start) > 1)
-        stop("start contains NAs and length(start) not 1", call.=FALSE)
-    if(any(is.na(count)) && length(count) > 1)
-        stop("count contains NAs and length(count) not 1", call.=FALSE)
+    stopifnot(length(start) == ndims)
+    start[is.na(start)] <- 1
 
-    if(any(is.na(start)) && varinfo$ndims > 0)
-        start <- rep(1, ndims)
-    if(any(is.na(count)) && varinfo$ndims > 0) {
-        if(is.null(dim(data)))
-	    count <- length(data)
-	else 
-	    count <- dim(data)
-
-        if(varinfo$type == "NC_CHAR")
-	    count <- c(max(nchar(data)), count)
+    if (isTRUE(is.na(count))) {
+      if (!is.null(dim(data))) {
+        count <- dim(data)
+      } else if (ndims==0 && length(data)==1) {
+        count <- integer(0)
+      } else {
+        count <- length(data)
+      }
+      if (is.character(data) && ndims > 0) {
+        strlen <- dim.inq.nc(ncfile, varinfo$dimids[1])$length
+        count <- c(strlen, count)
+      }    
+    } else if (length(count) > ndims) {
+      count <- count[seq_len(ndims)]
+    }
+    stopifnot(length(count) == ndims)
+    for (idim in seq_len(ndims)) {
+      if (is.na(count[idim])) {
+	diminfo <- dim.inq.nc(ncfile, varinfo$dimids[idim])
+	count[idim] <- ( diminfo$length - start[idim] + 1 )
+      }
     }
 
-    #-- Switch from R to C convention ------------------------------------------#
-    if(varinfo$ndims > 0) {
-	c.start <- start[length(start):1] - 1
-	c.count <- count[length(count):1]    
+    #-- Check that length of data is sufficient --------------------------------#
+    if (is.character(data) && ndims > 0) {
+      numelem <- prod(count[-1])
     } else {
-        c.start <- 0
-	c.count <- 1
-	
-	if(length(data) > 1)
-	    stop("Edge+start exceeds dimension bound", call.=FALSE)
+      numelem <- prod(count) # Returns 1 if ndims==0 (scalar variable)
+    }
+    if (length(data) < numelem) {
+      stop(paste("Not enough data elements (found ",length(data),
+             ", need ",numelem,")", sep=""), call.=FALSE)
     }
 
-    #-- Check for correct number of dimensions in start and count --------------#
-    if(varinfo$ndims > 0)
-	if(length(c.start) != ndims || length(c.count) != ndims)
-            stop("Length of start/count is not ndims", call.=FALSE)
-
-    #-- Check if length of data and count match (numeric data) -----------------#
-    if(is.numeric(data) && length(data) != prod(c.count))
-        stop("Mismatch in count and length(data)", call.=FALSE)
-
-    #-- Checks and definitions for character data ------------------------------#
-    if(is.character(data) && varinfo$ndims > 0) {
-        lastdim   <- varinfo$dimids[1]                       ## R to C convention
-	maxstrlen <- dim.inq.nc(ncfile, lastdim)$length
-    
-        if(length(data) != prod(c.count[1:(ndims-1)]))    ## already C convention
-	    stop("Mismatch in count and length(data)", call.=FALSE)
-	
-	if(max(nchar(data)) > maxstrlen)
-	    stop("max(nchar(data)) exceeding variable limits", call.=FALSE)
+    #-- Warn if strings will be truncated --------------------------------------#
+    if (is.character(data)) {
+      if (ndims > 0) {
+        strlen <- count[1]
+      } else {
+        strlen <- 1
+      }
+      if (max(nchar(data,type="bytes")) > strlen) {
+        warning(paste("Strings truncated to length",strlen), call.=FALSE)
+      }
     }
 
-    if(is.character(data)) {
-	varsize    <- vector()
-	varsize[1] <- prod(count)
-	varsize[2] <- count[1]
+    #-- Warn if array data is not conformable with count -----------------------#
+    if (!is.null(dim(data))) {
+      if (is.character(data)) {
+	count_drop <- count[-1]
+      } else {
+        count_drop <- count
+      }
+      count_drop <- count_drop[count_drop!=1]
+
+      dim_drop <- dim(data)
+      dim_drop <- dim_drop[dim_drop!=1]
+
+      if ((length(count_drop) != length(dim_drop)) || 
+	  any(count_drop != dim_drop)) {
+	warning(paste("Data coerced from dimensions (",
+                  paste(dim(data),collapse=","), ") to dimensions (",
+		  paste(count,collapse=","), ")", sep=""), call.=FALSE)
+      }
     }
 
     #-- Pack variables if requested (missing values are preserved) -------------#
-    if(pack) {
+    if(pack && is.numeric(data)) {
         offset <- try(att.inq.nc(ncfile, varinfo$name, "add_offset"),
 		      silent=TRUE)
         scale  <- try(att.inq.nc(ncfile, varinfo$name, "scale_factor"),
@@ -1085,70 +1069,77 @@ var.put.nc <- function(ncfile, variable, data, start=NA, count=NA, na.mode=0,
     }
 
     #-- Convert missing value to NA if defined in NetCDF file ------------------#
-    if(is.numeric(data)) {
-        if(any(is.na(data))) {
-	    na.flag <- 0
+    if(is.numeric(data) && any(is.na(data))) {
+      na.flag <- 0
 
-            missval.flag <- 0
-	    fillval.flag <- 0
+      missval.flag <- 0
+      fillval.flag <- 0
 
-            fillval <- try(att.inq.nc(ncfile, varinfo$name, "_FillValue"   ), 
-	        silent=TRUE)
-            missval <- try(att.inq.nc(ncfile, varinfo$name, "missing_value"),
-	        silent=TRUE)
+      fillval <- try(att.inq.nc(ncfile, varinfo$name, "_FillValue"   ), 
+	  silent=TRUE)
+      missval <- try(att.inq.nc(ncfile, varinfo$name, "missing_value"),
+	  silent=TRUE)
 
-            if(!(class(fillval) == "try-error"))
-		if(!is.null(fillval))
-	            fillval.flag <- 1
-	    if(!(class(missval) == "try-error"))
-		if(!is.null(missval))
-		    missval.flag <- 1
+      if(!(class(fillval) == "try-error"))
+	  if(!is.null(fillval))
+	      fillval.flag <- 1
+      if(!(class(missval) == "try-error"))
+	  if(!is.null(missval))
+	      missval.flag <- 1
 
-       	    if(na.mode == 0 && missval.flag == 1) {
-		na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
-		na.flag  <- 1
-	    }
-	    if(na.mode == 0 && fillval.flag == 1) {
-		na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
-		na.flag  <- 1
-	    }
+      if(na.mode == 0 && missval.flag == 1) {
+	  na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
+	  na.flag  <- 1
+      }
+      if(na.mode == 0 && fillval.flag == 1) {
+	  na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
+	  na.flag  <- 1
+      }
 
-       	    if(na.mode == 1 && fillval.flag == 1) {
-		na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
-		na.flag  <- 1
-	    }
+      if(na.mode == 1 && fillval.flag == 1) {
+	  na.value <- att.get.nc(ncfile, varinfo$name, "_FillValue")
+	  na.flag  <- 1
+      }
 
-	    if(na.mode == 2 && missval.flag == 1) {
-		na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
-		na.flag  <- 1
-	    }
+      if(na.mode == 2 && missval.flag == 1) {
+	  na.value <- att.get.nc(ncfile, varinfo$name, "missing_value")
+	  na.flag  <- 1
+      }
 
-            if(na.flag == 1)
-		data[is.na(data)] <- na.value
-	    else
-	        stop("Found NAs but no missing value attribute", call.=FALSE)
-        }
+      if(na.flag == 1)
+	  data[is.na(data)] <- as.numeric(na.value)
+      else
+	  stop("Found NAs but no missing value attribute", call.=FALSE)
     }
 
+    #-- Switch from R to C convention ------------------------------------------#
+    c.start <- rev(start-1)
+    c.count <- rev(count)
+
     #-- C function calls -------------------------------------------------------#
-    if(is.numeric(data))
-	nc <- .Call("R_nc_put_vara_double",
+    if(is.numeric(data)) {
+        if (!is.double(data)) {
+          data <- as.double(data)
+        }
+	nc <- .Call(R_nc_put_vara_double,
         	    as.integer(ncfile),
 		    as.integer(varid),
 		    as.integer(c.start),
 		    as.integer(c.count),
-		    as.double(data),
-		    PACKAGE="RNetCDF")
-    else
-        nc <- .Call("R_nc_put_vara_text",
+		    as.integer(ndims),
+		    data)
+    } else {
+        stopifnot(is.character(data) || is.raw(data))
+        nc <- .Call(R_nc_put_vara_text,
         	    as.integer(ncfile),
 		    as.integer(varid),
 		    as.integer(c.start),
 		    as.integer(c.count),
-		    as.character(data),
-		    as.integer(varsize),
-		    PACKAGE="RNetCDF")
-		    
+		    as.integer(ndims),
+                    as.integer(is.raw(data)),
+		    data)
+    }
+
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
 }
@@ -1173,13 +1164,12 @@ var.rename.nc <- function(ncfile, variable, newname)
     ifelse(is.character(variable), varname <- variable, varid <- variable)
 
     #-- C function call --------------------------------------------------------#
-    nc <- .Call("R_nc_rename_var",
+    nc <- .Call(R_nc_rename_var,
         	as.integer(ncfile),
 		as.integer(varid),
 		as.character(varname),
 		as.integer(nameflag),
-		as.character(newname),
-		PACKAGE="RNetCDF")
+		as.character(newname))
 
     if(nc$status != 0)
         stop(nc$errmsg, call.=FALSE)
@@ -1227,32 +1217,37 @@ utcal.nc <- function(unitstring, value, type="n")
     #-- Check args -------------------------------------------------------------#
     stopifnot(is.character(unitstring))
     stopifnot(is.numeric(value) && !any(is.na(value)))
-    stopifnot(type == "n" || type =="s")
+    stopifnot(type == "n" || type =="s" || type == "c" )
     
     count <- length(value)
-    
-    #-- C function call --------------------------------------------------------#
-    ut <- .Call("R_ut_calendar", 
-		as.character(unitstring), 
+   
+    #-- C function call to udunits calendar function -----------------------#
+    ut <- .Call(R_ut_calendar, 
+	        as.character(unitstring), 
 		as.integer(count),
-		as.double(value),
-		PACKAGE="RNetCDF")
+		as.double(value))
 
-    #-- Return object if no error ----------------------------------------------#
+    #-- Return object if no error ------------------------------------------#
     if(ut$status == 0) {
-        if(type == "n") {
+	if(type == "n") {
 	    colnames(ut$value) <- c("year", "month", "day", "hour", 
 		"minute", "second")
-            return(ut$value)
-        } else {
+	    return(ut$value)
+	} else if (type == "s") {
 	    x <- apply(ut$value, 1, function(x){paste(x[1],"-",
-	               sprintf("%02g",x[2]),"-",sprintf("%02g",x[3])," ",
+		       sprintf("%02g",x[2]),"-",sprintf("%02g",x[3])," ",
 		       sprintf("%02g",x[4]),":",sprintf("%02g",x[5]),":",
 		       sprintf("%02g",x[6]),sep="")})
 	    return(x)
-	}
-    } else
-        stop(ut$errmsg, call.=FALSE)
+        } else if (type == "c") {
+            ct <- as.POSIXct(
+                        utinvcal.nc("seconds since 1970-01-01 00:00:00 +00:00",ut$value),
+                        tz="UTC", origin=ISOdatetime(1970,1,1,0,0,0,tz="UTC"))
+            return(ct)
+        }
+    } else {
+	stop(ut$errmsg, call.=FALSE)
+    }
 }
 
 
@@ -1262,9 +1257,8 @@ utcal.nc <- function(unitstring, value, type="n")
 
 utinit.nc <- function(path="")
 {
-    ut <- .Call("R_ut_init", 
-                as.character(path),
-		PACKAGE="RNetCDF")
+    ut <- .Call(R_ut_init, 
+                as.character(path))
 		
     if(ut$status != 0)
         stop(ut$errmsg, call.=FALSE)
@@ -1279,17 +1273,20 @@ utinvcal.nc <- function(unitstring, value)
 {
     #-- Check args -------------------------------------------------------------#
     stopifnot(is.character(unitstring))
-    
-    if(is.character(value)) {
-	stopifnot(any(nchar(value) == 19))
+
+    if (is.character(value)) {
+	stopifnot(all(nchar(value) == 19))
 	value <- cbind(substr(value,1,4),
-                       substr(value,6,7),
+		       substr(value,6,7),
 		       substr(value,9,10),
 		       substr(value,12,13),
 		       substr(value,15,16),
 		       substr(value,18,19))
 
 	value <- matrix(as.numeric(value),ncol=6)
+    } else if (inherits(value,"POSIXct")) {
+        value <- utcal.nc("seconds since 1970-01-01 00:00:00 +00:00",
+                     as.numeric(value), 'n')
     }
 
     stopifnot(is.numeric(value) && !any(is.na(value)))
@@ -1300,20 +1297,20 @@ utinvcal.nc <- function(unitstring, value)
 	stop("length(value) not divisible by 6", call.=FALSE)
 
     if(is.matrix(value) && ncol(value) != 6) 
-        stop("ncol(value) not 6", call.=FALSE)
+	stop("ncol(value) not 6", call.=FALSE)
     
     #-- C function call --------------------------------------------------------#
-    ut <- .Call("R_ut_inv_calendar", 
+    ut <- .Call(R_ut_inv_calendar, 
 		as.character(unitstring), 
 		as.integer(count),
-		as.double(value),
-		PACKAGE="RNetCDF")
+		as.double(value))
 
     #-- Return object if no error ----------------------------------------------#
-    if(ut$status == 0)
-        return(ut$value)
-    else
-        stop(ut$errmsg, call.=FALSE)
+    if(ut$status == 0) {
+	return(ut$value)
+    } else {
+	stop(ut$errmsg, call.=FALSE)
+    }
 }
 
 
