@@ -2,14 +2,14 @@
  *									       *
  *  Name:       RNetCDF.c						       *
  *									       *
- *  Version:    1.6.3-1							       *
+ *  Version:    1.9-1                                                          *
  *									       *
  *  Purpose:    NetCDF interface for R.					       *
  *									       *
- *  Author:     Pavel Michna (michna@giub.unibe.ch)			       *
- *              Milton Woods (m.woods@bom.gov.au)                              *
+ *  Author:     Pavel Michna (michna@users.r-forge.r-project.org)	       *
+ *              Milton Woods (mwoods@users.r-forge.r-project.org)              *
  *									       *
- *  Copyright:  (C) 2004-2014 Pavel Michna                                     *
+ *  Copyright:  (C) 2004-2017 Pavel Michna                                     *
  *									       *
  *=============================================================================*
  *									       *
@@ -50,6 +50,16 @@
  *  pm       04/01/11   Corrected string handling in R_nc_get_vara_text        *
  *  pm       05/01/11   Removed extra zeroing after Calloc                     *
  *  pm       26/05/14   Corrected memory leak issue (lines 1338 and 1593)      *
+ *  mw       05/09/14   Support reading and writing raw character arrays,      *
+ *                      avoid temporary arrays when reading/writing variables  *
+ *  mw       08/09/14   Handle reading and writing of zero-sized arrays        *
+ *  mw       01/02/15   Remove redundant ut_read_xml from R_ut_init            *
+ *  mw       24/04/15   Initialise and free utunit when using udunits2,        *
+ *                      to fix memory errors reported by valgrind.             *
+ *                      Allow udunits2 headers to be in udunits2 directory.    *
+ *  mw       26/01/16   Fix memory leak from abnormal exit of calendar funcs.  *
+ *  mw       16/07/17   Minimise use of nc_enddef to read data from opendap.   *
+ *  mw       01/10/17   Register native routines for invocation by R           *
  *									       *
 \*=============================================================================*/
 
@@ -62,10 +72,32 @@
 #include <string.h>
 
 #include <netcdf.h>
-#include <udunits.h>
+
+#ifdef HAVE_UDUNITS2_UDUNITS_H
+  #include <udunits2/udunits.h>
+#else
+  #include <udunits.h>
+#endif
 
 #include <R.h>
 #include <Rinternals.h>
+#include <R_ext/Rdynload.h>
+
+
+/*=============================================================================*\
+ *  Macros								       *
+\*=============================================================================*/
+
+/* Execute a netcdf function, changing to netcdf data mode if necessary.
+ * Function return value is stored in parameter status. */
+#define NC_DATAMODE(status, ncid, cmd) \
+  status = cmd; \
+  if (status == NC_EINDEFINE) { \
+    status = nc_enddef(ncid); \
+    if (status == NC_NOERR) { \
+      status = cmd; \
+    } \
+  }
 
 
 /*=============================================================================*\
@@ -1045,7 +1077,7 @@ SEXP R_nc_open (SEXP filename, SEXP write, SEXP share, SEXP prefill)
 
 SEXP R_nc_sync (SEXP ncid)
 {
-    int  status;
+    int  status, incid;
     SEXP retlist, retlistnames;
 
     /*-- Create output object and initialize return values --------------------*/
@@ -1062,19 +1094,12 @@ SEXP R_nc_sync (SEXP ncid)
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     SET_VECTOR_ELT (retlist, 1, mkString(""));
 
-    /*-- Enter data mode (if necessary) ---------------------------------------*/
-    status = nc_enddef(INTEGER(ncid)[0]);
-    if((status != NC_NOERR) && (status != NC_ENOTINDEFINE)) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	return(retlist);
-    }
-
     /*-- Sync the file --------------------------------------------------------*/
-    status = nc_sync(INTEGER(ncid)[0]);
-    if(status != NC_NOERR)
+    incid = INTEGER(ncid)[0];
+    NC_DATAMODE(status, incid, nc_sync(incid));
+    if(status != NC_NOERR) {
 	SET_VECTOR_ELT(retlist, 1, mkString(nc_strerror(status)));
+    }
 
     /*-- Returning the list ---------------------------------------------------*/
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;
@@ -1173,18 +1198,25 @@ SEXP R_nc_def_var (SEXP ncid, SEXP varname, SEXP type, SEXP ndims, SEXP dimids)
 \*-----------------------------------------------------------------------------*/
 
 SEXP R_nc_get_vara_double (SEXP ncid, SEXP varid, SEXP start, 
-                           SEXP count, SEXP varsize)
+                           SEXP count, SEXP ndims)
 {
-    int    ndims, i, status;
-    double *data;
-    size_t s_start[MAX_NC_DIMS], s_count[MAX_NC_DIMS];
+    int    i, status, incid;
+    size_t s_start[MAX_NC_DIMS], s_count[MAX_NC_DIMS], varsize;
     SEXP   retlist, retlistnames;
 
+    /*-- Copy dims from int to size_t, calculate total array size -------------*/
+    varsize = 1;
+    for(i=0; i<INTEGER(ndims)[0]; i++) {
+	s_start[i] = (size_t)INTEGER(start)[i];
+	s_count[i] = (size_t)INTEGER(count)[i];
+	varsize *= s_count[i];
+    }
+	
     /*-- Create output object and initialize return values --------------------*/
     PROTECT(retlist = allocVector(VECSXP, 3));
     SET_VECTOR_ELT(retlist, 0, allocVector(REALSXP, 1));
     SET_VECTOR_ELT(retlist, 1, allocVector(STRSXP,  1));
-    SET_VECTOR_ELT(retlist, 2, allocVector(REALSXP, INTEGER(varsize)[0]));
+    SET_VECTOR_ELT(retlist, 2, allocVector(REALSXP, varsize));
 
     PROTECT(retlistnames = allocVector(STRSXP, 3)); 
     SET_STRING_ELT(retlistnames, 0, mkChar("status")); 
@@ -1192,62 +1224,25 @@ SEXP R_nc_get_vara_double (SEXP ncid, SEXP varid, SEXP start,
     SET_STRING_ELT(retlistnames, 2, mkChar("data")); 
     setAttrib(retlist, R_NamesSymbol, retlistnames); 
 
-    data = Calloc(INTEGER(varsize)[0], double);
-
     status = -1;
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     SET_VECTOR_ELT (retlist, 1, mkString(""));
 
-    /*-- Get ndims for this var -----------------------------------------------*/
-    status = nc_inq_varndims(INTEGER(ncid)[0], INTEGER(varid)[0], &ndims);
-    if(status != NC_NOERR) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-	REAL(VECTOR_ELT(retlist, 0))[0] = status;
-        UNPROTECT(2);
-	Free(data);
-	return(retlist);
-    }
-
-    /*-- Copy over from int to size_t, handle scalar variables ----------------*/
-    if(ndims > 0) {
-	for(i=0; i<ndims; i++) {
-	    s_start[i] = (size_t)INTEGER(start)[i];
-	    s_count[i] = (size_t)INTEGER(count)[i];
-	}
-    }
-    else {
-        s_start[0] = 0;
-	s_count[0] = 1;
-    }
-		
-    /*-- Enter data mode (if necessary) ---------------------------------------*/
-    status = nc_enddef(INTEGER(ncid)[0]);
-    if((status != NC_NOERR) && (status != NC_ENOTINDEFINE)) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	Free(data);
-	return(retlist);
-    }
-
-    /*-- Get the var ----------------------------------------------------------*/
-    status = nc_get_vara_double(INTEGER(ncid)[0], INTEGER(varid)[0],
-        s_start, s_count, data);
-    if(status != NC_NOERR) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-	REAL(VECTOR_ELT(retlist, 0))[0] = status;
-        UNPROTECT(2);
-	Free(data);
-	return(retlist);
+    /*-- Read variable from file ----------------------------------------------*/
+    if (varsize > 0) {
+      /* Some netcdf versions cannot handle zero-sized arrays */
+      incid = INTEGER(ncid)[0];
+      NC_DATAMODE(status, incid, \
+        nc_get_vara_double(incid, INTEGER(varid)[0], \
+        s_start, s_count, REAL(VECTOR_ELT(retlist, 2))));
+    } else {
+      status = NC_NOERR;
     }
    
-    /*-- Copy from C to R object ----------------------------------------------*/
-    for(i=0; i<INTEGER(varsize)[0]; i++)
-        REAL(VECTOR_ELT(retlist, 2))[i] = (double)data[i];
-
-    Free(data);
-
     /*-- Returning the list ---------------------------------------------------*/
+    if(status != NC_NOERR) {
+      SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
+    }
     REAL(VECTOR_ELT(retlist, 0))[0] = status;
     UNPROTECT(2);
     return(retlist);
@@ -1259,21 +1254,42 @@ SEXP R_nc_get_vara_double (SEXP ncid, SEXP varid, SEXP start,
 \*-----------------------------------------------------------------------------*/
 
 SEXP R_nc_get_vara_text (SEXP ncid, SEXP varid, SEXP start, 
-                         SEXP count, SEXP varsize)
+                         SEXP count, SEXP ndims, SEXP rawchar)
 {
-    int    ndims, tx_len, tx_num, i, j, status;
+    int    i, status, incid;
     char   *data, *tx_str;
     size_t s_start[MAX_NC_DIMS], s_count[MAX_NC_DIMS];
+    size_t tx_len, tx_num, varsize;
     SEXP   retlist, retlistnames;
 
+    /*-- Copy dims from int to size_t, calculate number and length of strings -*/
+    for(i=0; i<INTEGER(ndims)[0]; i++) {
+	s_start[i] = (size_t)INTEGER(start)[i];
+	s_count[i] = (size_t)INTEGER(count)[i];
+    }
+
+    if(INTEGER(ndims)[0] > 0) {
+        tx_num = 1;
+	for(i=0; i<INTEGER(ndims)[0]-1; i++) {
+            tx_num *= s_count[i];
+	}
+        tx_len = s_count[INTEGER(ndims)[0]-1];
+    } else {
+        tx_num = 1;
+        tx_len = 1;
+    }
+    varsize = tx_num*tx_len;
+
     /*-- Create output object and initialize return values --------------------*/
-    tx_len = INTEGER(varsize)[1];
-    tx_num = INTEGER(varsize)[0]/tx_len;
-    
     PROTECT(retlist = allocVector(VECSXP, 3));
     SET_VECTOR_ELT(retlist, 0, allocVector(REALSXP, 1));
     SET_VECTOR_ELT(retlist, 1, allocVector(STRSXP,  1));
-    SET_VECTOR_ELT(retlist, 2, allocVector(STRSXP,  tx_num));
+
+    if (INTEGER(rawchar)[0] > 0) {
+      SET_VECTOR_ELT(retlist, 2, allocVector(RAWSXP, varsize));
+    } else {
+      SET_VECTOR_ELT(retlist, 2, allocVector(STRSXP, tx_num));
+    }
 
     PROTECT(retlistnames = allocVector(STRSXP, 3)); 
     SET_STRING_ELT(retlistnames, 0, mkChar("status")); 
@@ -1285,64 +1301,37 @@ SEXP R_nc_get_vara_text (SEXP ncid, SEXP varid, SEXP start,
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     SET_VECTOR_ELT (retlist, 1, mkString(""));
 
-    data = Calloc(INTEGER(varsize)[0], char);                 /*-- Is zeroed --*/
-
-    /*-- Get ndims for this var -----------------------------------------------*/
-    status = nc_inq_varndims(INTEGER(ncid)[0], INTEGER(varid)[0], &ndims);
-    if(status != NC_NOERR) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-	REAL(VECTOR_ELT(retlist, 0))[0] = status;
-        UNPROTECT(2);
-	Free(data);
-	return(retlist);
+    /*-- Read variable from file ----------------------------------------------*/
+    if (INTEGER(rawchar)[0] > 0) {
+      data = (char *) RAW(VECTOR_ELT(retlist, 2));
+    } else {
+      data = (char *) R_alloc(varsize, sizeof(char));
     }
 
-    /*-- Copy over from int to size_t, handle scalar variables ----------------*/
-    if(ndims > 0) {
-	for(i=0; i<ndims; i++) {
-	    s_start[i] = (size_t)INTEGER(start)[i];
-	    s_count[i] = (size_t)INTEGER(count)[i];
-	}
-    }
-    else {
-        s_start[0] = 0;
-	s_count[0] = 1;
+    if (varsize > 0) {
+      /* Some netcdf versions cannot handle zero-sized arrays */
+      incid = INTEGER(ncid)[0];
+      NC_DATAMODE(status, incid, \
+        nc_get_vara_text(incid, INTEGER(varid)[0], \
+        s_start, s_count, data));
+    } else {
+      status = NC_NOERR;
     }
 
-    /*-- Enter data mode (if necessary) ---------------------------------------*/
-    status = nc_enddef(INTEGER(ncid)[0]);
-    if((status != NC_NOERR) && (status != NC_ENOTINDEFINE)) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	Free(data);
-	return(retlist);
-    }
-
-    /*-- Get the var ----------------------------------------------------------*/
-    status = nc_get_vara_text(INTEGER(ncid)[0], INTEGER(varid)[0],
-        s_start, s_count, data);
-    if(status != NC_NOERR) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-	REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;
-        UNPROTECT(2);
-	Free(data);
-	return(retlist);
-    }
-   
-    /*-- Copy from C to R object ----------------------------------------------*/
-    tx_str = Calloc(tx_len+1, char);                    /*-- String handling --*/
-    for(i=0; i<tx_num; i++) {
-        for(j=0; j<tx_len; j++)
-            tx_str[j] = data[i*tx_len+j];
-	tx_str[j] = '\0';                               /*-- String handling --*/
+    /*-- Copy from C to R character vector (if specified) ---------------------*/
+    if (status == NC_NOERR && INTEGER(rawchar)[0] <= 0) {
+      tx_str = (char *) R_alloc(tx_len+1, sizeof(char));
+      tx_str[tx_len] = '\0'; /* Final null character is never modified */
+      for(i=0; i<tx_num; i++) {
+        strncpy(tx_str, data+i*tx_len, tx_len);
 	SET_STRING_ELT(VECTOR_ELT(retlist, 2), i, mkChar(tx_str));
+      }
     }
 
-    Free(data);
-    Free(tx_str);
-
     /*-- Returning the list ---------------------------------------------------*/
+    if (status != NC_NOERR) {
+      SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
+    }
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;
     UNPROTECT(2);
     return(retlist);
@@ -1475,12 +1464,20 @@ SEXP R_nc_inq_var (SEXP ncid, SEXP varid, SEXP varname, SEXP nameflag)
 \*-----------------------------------------------------------------------------*/
 
 SEXP R_nc_put_vara_double (SEXP ncid, SEXP varid, SEXP start, 
-                           SEXP count, SEXP data)
+                           SEXP count, SEXP ndims, SEXP data)
 {
-    int    ndims, i, status;
-    size_t s_start[MAX_NC_DIMS], s_count[MAX_NC_DIMS];
+    int    i, status, incid;
+    size_t s_start[MAX_NC_DIMS], s_count[MAX_NC_DIMS], varsize;
     SEXP   retlist, retlistnames;
 
+    /*-- Copy dims from int to size_t -----------------------------------------*/
+    varsize = 1;
+    for(i=0; i<INTEGER(ndims)[0]; i++) {
+	s_start[i] = (size_t)INTEGER(start)[i];
+	s_count[i] = (size_t)INTEGER(count)[i];
+        varsize *= s_count[i];
+    }
+
     /*-- Create output object and initialize return values --------------------*/
     PROTECT(retlist = allocVector(VECSXP, 2));
     SET_VECTOR_ELT(retlist, 0, allocVector(REALSXP, 1));
@@ -1495,43 +1492,21 @@ SEXP R_nc_put_vara_double (SEXP ncid, SEXP varid, SEXP start,
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     SET_VECTOR_ELT (retlist, 1, mkString(""));
 
-    /*-- Get ndims for this var -----------------------------------------------*/
-    status = nc_inq_varndims(INTEGER(ncid)[0], INTEGER(varid)[0], &ndims);
-    if(status != NC_NOERR) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;	 
-	UNPROTECT(2);
-	return(retlist);
-    }
-
-    /*-- Copy over from int to size_t, handle scalar variables ----------------*/
-    if(ndims > 0) {
-	for(i=0; i<ndims; i++) {
-	    s_start[i] = (size_t)INTEGER(start)[i];
-	    s_count[i] = (size_t)INTEGER(count)[i];
-	}
-    }
-    else {
-        s_start[0] = 0;
-	s_count[0] = 1;
-    }
-
-    /*-- Enter data mode (if necessary) ---------------------------------------*/
-    status = nc_enddef(INTEGER(ncid)[0]);
-    if((status != NC_NOERR) && (status != NC_ENOTINDEFINE)) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	return(retlist);
-    }
-
     /*-- Put the var ----------------------------------------------------------*/
-    status = nc_put_vara_double(INTEGER(ncid)[0], INTEGER(varid)[0],
-        s_start, s_count, REAL(data));
-    if(status != NC_NOERR)
-        SET_VECTOR_ELT(retlist, 1, mkString(nc_strerror(status)));
+    if (varsize > 0) {
+      /* Some netcdf versions cannot handle zero-sized arrays */
+      incid = INTEGER(ncid)[0];
+      NC_DATAMODE(status, incid, \
+        nc_put_vara_double(incid, INTEGER(varid)[0], \
+        s_start, s_count, REAL(data)));
+    } else {
+      status = NC_NOERR;
+    }
 
     /*-- Returning the list ---------------------------------------------------*/
+    if (status != NC_NOERR) {
+      SET_VECTOR_ELT(retlist, 1, mkString(nc_strerror(status)));
+    }
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     UNPROTECT(2);
     return(retlist);
@@ -1543,17 +1518,33 @@ SEXP R_nc_put_vara_double (SEXP ncid, SEXP varid, SEXP start,
 \*-----------------------------------------------------------------------------*/
 
 SEXP R_nc_put_vara_text (SEXP ncid, SEXP varid, SEXP start, 
-                         SEXP count, SEXP data, SEXP varsize)
+                         SEXP count, SEXP ndims, SEXP rawchar, SEXP data)
 {
-    int    ndims, tx_num, tx_len, i, j, status;
-    char   *ncdata, *tx_str;
+    int    i, status, incid;
+    char   *ncdata;
     size_t s_start[MAX_NC_DIMS], s_count[MAX_NC_DIMS];
+    size_t tx_len, tx_num, varsize;
     SEXP   retlist, retlistnames;
 
+    /*-- Copy dims from int to size_t, calculate number and length of strings -*/
+    for(i=0; i<INTEGER(ndims)[0]; i++) {
+	s_start[i] = (size_t)INTEGER(start)[i];
+	s_count[i] = (size_t)INTEGER(count)[i];
+    }
+
+    if (INTEGER(ndims)[0] > 0) {
+	tx_num = 1;
+	for(i=0; i<INTEGER(ndims)[0]-1; i++) {
+	    tx_num *= s_count[i];
+	}
+	tx_len = s_count[INTEGER(ndims)[0]-1];
+    } else {
+	tx_num = 1;
+	tx_len = 1;
+    }
+    varsize = tx_num*tx_len;
+
     /*-- Create output object and initialize return values --------------------*/
-    tx_len = INTEGER(varsize)[1];
-    tx_num = INTEGER(varsize)[0]/tx_len;
-    
     PROTECT(retlist = allocVector(VECSXP, 2));
     SET_VECTOR_ELT(retlist, 0, allocVector(REALSXP, 1));
     SET_VECTOR_ELT(retlist, 1, allocVector(STRSXP,  1));
@@ -1567,60 +1558,31 @@ SEXP R_nc_put_vara_text (SEXP ncid, SEXP varid, SEXP start,
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     SET_VECTOR_ELT (retlist, 1, mkString(""));
 
-    /*-- Get ndims for this var -----------------------------------------------*/
-    status = nc_inq_varndims(INTEGER(ncid)[0], INTEGER(varid)[0], &ndims);
-    if(status != NC_NOERR) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;	 
-	UNPROTECT(2);
-	return(retlist);
+    /*-- Prepare output array -------------------------------------------------*/
+    if (INTEGER(rawchar)[0] > 0) {
+      ncdata = (char *) RAW(data);
+    } else {
+      ncdata = (char *) R_alloc(varsize, sizeof(char));
+      for(i=0; i<tx_num; i++) {
+	  strncpy(ncdata+i*tx_len, CHAR(STRING_ELT(data, i)), tx_len);
+      }
     }
-
-    /*-- Copy over from int to size_t, handle scalar variables ----------------*/
-    if(ndims > 0) {
-	for(i=0; i<ndims; i++) {
-	    s_start[i] = (size_t)INTEGER(start)[i];
-	    s_count[i] = (size_t)INTEGER(count)[i];
-	}
+ 
+    /*-- Write variable to file -----------------------------------------------*/
+    if (varsize > 0) {
+      /* Some netcdf versions cannot handle zero-sized arrays */
+      incid = INTEGER(ncid)[0];
+      NC_DATAMODE(status, incid, \
+        nc_put_vara_text(incid, INTEGER(varid)[0], \
+	  s_start, s_count, ncdata));
+    } else {
+      status = NC_NOERR;
     }
-    else {
-        s_start[0] = 0;
-	s_count[0] = 1;
-    }
-
-    /*-- Copy from R to C object ----------------------------------------------*/
-    ncdata = Calloc(tx_len*tx_num, char);                     /*-- Is zeroed --*/
-    tx_str = Calloc(tx_len+1, char);                          /*-- Is zeroed --*/
-        
-    for(i=0; i<tx_num; i++) {
-        strcpy(tx_str, CHAR(STRING_ELT(data, i)));
-	for(j=0; j<tx_len; j++) {
-            ncdata[i*tx_len+j] = tx_str[j];
-	    tx_str[j] = '\0';
-	}
-    }
-
-    /*-- Enter data mode (if necessary) ---------------------------------------*/
-    status = nc_enddef(INTEGER(ncid)[0]);
-    if((status != NC_NOERR) && (status != NC_ENOTINDEFINE)) {
-        SET_VECTOR_ELT (retlist, 1, mkString(nc_strerror(status)));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	Free(ncdata);
-	Free(tx_str);
-	return(retlist);
-    }
-
-    /*-- Put the var ----------------------------------------------------------*/
-    status = nc_put_vara_text(INTEGER(ncid)[0], INTEGER(varid)[0],
-        s_start, s_count, ncdata);
-    if(status != NC_NOERR)
-        SET_VECTOR_ELT(retlist, 1, mkString(nc_strerror(status)));
-
-    Free(ncdata);
-    Free(tx_str);
 
     /*-- Returning the list ---------------------------------------------------*/
+    if(status != NC_NOERR) {
+      SET_VECTOR_ELT(retlist, 1, mkString(nc_strerror(status)));
+    }
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	 
     UNPROTECT(2);
     return(retlist);
@@ -1769,32 +1731,24 @@ SEXP R_ut_calendar (SEXP unitstring, SEXP unitcount, SEXP values)
     SET_VECTOR_ELT (retlist, 1, mkString(""));
     
     /*-- Scan unitstring ------------------------------------------------------*/
+#ifdef HAVE_LIBUDUNITS2
+    utIni(&utunit);
+#endif
+
     status = utScan(CHAR(STRING_ELT(unitstring, 0)), &utunit);
     if(status != 0) {
-        R_ut_strerror(status, strerror);
-        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	return(retlist);        
+        goto cleanup;
     }
   
     /*-- Check if unit is time and has origin ---------------------------------*/
-    status = utIsTime(&utunit);    
-    if(status == 0) {
-        R_ut_strerror(UT_ENOTTIME, strerror);
-        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
-	REAL(VECTOR_ELT(retlist, 0))[0] = UT_ENOTTIME;
-	UNPROTECT(2);
-	return(retlist);        
+    if ( !utIsTime(&utunit) ) {
+        status = UT_ENOTTIME;
+        goto cleanup;
     }
 
-    status = utHasOrigin(&utunit);    
-    if(status == 0) {
-        R_ut_strerror(UT_EINVALID, strerror);
-        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
-        REAL(VECTOR_ELT(retlist, 0))[0] = UT_EINVALID;
-	UNPROTECT(2);
-	return(retlist);        
+    if ( !utHasOrigin(&utunit) ) {
+        status = UT_EINVALID;
+        goto cleanup;
     }
 
     /*-- Convert values -------------------------------------------------------*/
@@ -1813,12 +1767,15 @@ SEXP R_ut_calendar (SEXP unitstring, SEXP unitcount, SEXP values)
     }
 
     /*-- Returning the list ---------------------------------------------------*/
+cleanup:
+#ifdef HAVE_LIBUDUNITS2
+    utFree(&utunit);
+#endif
+    REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	     
     if(status != 0) {
         R_ut_strerror(status, strerror);
-        SET_VECTOR_ELT(retlist, 1, mkString(strerror));
+        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
     }
-    
-    REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	     
     UNPROTECT(2);
     return(retlist);
 }
@@ -1835,12 +1792,9 @@ SEXP R_ut_init (SEXP path)
     SEXP  retlist, retlistnames;
 
     /*-- Avoid "overriding default" messages from UDUNITS-2 (1/2) -------------*/
-    #ifdef UT_UNITS2_H_INCLUDED
-        ut_system* unitSystem;
-
-        ut_set_error_message_handler(ut_ignore);
-        unitSystem = ut_read_xml(NULL);
-    #endif
+#ifdef HAVE_LIBUDUNITS2
+    ut_set_error_message_handler(ut_ignore);
+#endif
 
     /*-- Create output object and initialize return values --------------------*/
     PROTECT(retlist = allocVector(VECSXP, 2));
@@ -1864,9 +1818,9 @@ SEXP R_ut_init (SEXP path)
     }
 
     /*-- Avoid "overriding default" messages from UDUNITS-2 (2/2) -------------*/
-    #ifdef UT_UNITS2_H_INCLUDED
-        ut_set_error_message_handler(ut_write_to_stderr);
-    #endif
+#ifdef HAVE_LIBUDUNITS2
+    ut_set_error_message_handler(ut_write_to_stderr);
+#endif
 
     /*-- Returning the list ---------------------------------------------------*/
     REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;
@@ -1908,32 +1862,24 @@ SEXP R_ut_inv_calendar (SEXP unitstring, SEXP unitcount, SEXP values)
     SET_VECTOR_ELT (retlist, 1, mkString(""));
     
     /*-- Scan unitstring ------------------------------------------------------*/
+#ifdef HAVE_LIBUDUNITS2
+    utIni(&utunit);
+#endif
+
     status = utScan(CHAR(STRING_ELT(unitstring, 0)), &utunit);
     if(status != 0) {
-        R_ut_strerror(status, strerror);
-        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
-        REAL(VECTOR_ELT(retlist, 0))[0] = status;
-	UNPROTECT(2);
-	return(retlist);        
+        goto cleanup;
     }
 
     /*-- Check if unit is time and has origin ---------------------------------*/
-    status = utIsTime(&utunit);    
-    if(status == 0) {
-        R_ut_strerror(UT_ENOTTIME, strerror);
-        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
-        REAL(VECTOR_ELT(retlist, 0))[0] = UT_ENOTTIME;
-	UNPROTECT(2);
-	return(retlist);        
+    if ( !utIsTime(&utunit) ) {
+        status = UT_ENOTTIME;
+        goto cleanup;
     }
 
-    status = utHasOrigin(&utunit);    
-    if(status == 0) {
-        R_ut_strerror(UT_EINVALID, strerror);
-        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
-        REAL(VECTOR_ELT(retlist, 0))[0] = UT_EINVALID;
-	UNPROTECT(2);
-	return(retlist);        
+    if ( !utHasOrigin(&utunit) ) {
+        status = UT_EINVALID;
+        goto cleanup;
     }
   
     /*-- Convert values -------------------------------------------------------*/
@@ -1952,19 +1898,61 @@ SEXP R_ut_inv_calendar (SEXP unitstring, SEXP unitcount, SEXP values)
     }
 
     /*-- Returning the list ---------------------------------------------------*/
+cleanup:
+#ifdef HAVE_LIBUDUNITS2
+    utFree(&utunit);
+#endif
+    REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	     
     if(status != 0) {
         R_ut_strerror(status, strerror);
-        SET_VECTOR_ELT(retlist, 1, mkString(strerror));
+        SET_VECTOR_ELT (retlist, 1, mkString(strerror));
     }
-    
-    REAL(VECTOR_ELT(retlist, 0))[0] = (double)status;	     
     UNPROTECT(2);
     return(retlist);
 }
 
 
-/*=============================================================================*/
- 
+/*=============================================================================*\
+ *  Register native routines						       *
+\*=============================================================================*/
+
+static const R_CallMethodDef callMethods[]  = {
+  {"R_nc_copy_att", (DL_FUNC) &R_nc_copy_att, 7},
+  {"R_nc_delete_att", (DL_FUNC) &R_nc_delete_att, 4},
+  {"R_nc_get_att", (DL_FUNC) &R_nc_get_att, 5},
+  {"R_nc_inq_att", (DL_FUNC) &R_nc_inq_att, 6},
+  {"R_nc_put_att", (DL_FUNC) &R_nc_put_att, 8},
+  {"R_nc_rename_att", (DL_FUNC) &R_nc_rename_att, 5},
+  {"R_nc_close", (DL_FUNC) &R_nc_close, 1},
+  {"R_nc_create", (DL_FUNC) &R_nc_create, 5},
+  {"R_nc_def_dim", (DL_FUNC) &R_nc_def_dim, 4},
+  {"R_nc_inq_dim", (DL_FUNC) &R_nc_inq_dim, 4},
+  {"R_nc_rename_dim", (DL_FUNC) &R_nc_rename_dim, 5},
+  {"R_nc_inq_file", (DL_FUNC) &R_nc_inq_file, 1},
+  {"R_nc_open", (DL_FUNC) &R_nc_open, 4},
+  {"R_nc_sync", (DL_FUNC) &R_nc_sync, 1},
+  {"R_nc_def_var", (DL_FUNC) &R_nc_def_var, 5},
+  {"R_nc_get_vara_double", (DL_FUNC) &R_nc_get_vara_double, 5},
+  {"R_nc_get_vara_text", (DL_FUNC) &R_nc_get_vara_text, 6},
+  {"R_nc_inq_var", (DL_FUNC) &R_nc_inq_var, 4},
+  {"R_nc_put_vara_double", (DL_FUNC) &R_nc_put_vara_double, 6},
+  {"R_nc_put_vara_text", (DL_FUNC) &R_nc_put_vara_text, 7},
+  {"R_nc_rename_var", (DL_FUNC) &R_nc_rename_var, 5},
+  {"R_ut_calendar", (DL_FUNC) &R_ut_calendar, 3},
+  {"R_ut_init", (DL_FUNC) &R_ut_init, 1},
+  {"R_ut_inv_calendar", (DL_FUNC) &R_ut_inv_calendar, 3},
+  {NULL, NULL, 0}
+};
+
+void R_init_RNetCDF(DllInfo *info)
+{
+   R_registerRoutines(info, NULL, callMethods, NULL, NULL);
+   R_useDynamicSymbols(info, FALSE);
+   /* Omit R_forceSymbols for compatibility with R<3.0.0, and because
+      some users have reported that they call the C interface directly. */
+}
+
+
 /*=============================================================================*\
  *  SCRATCH                                                                    *
 \*=============================================================================*/
